@@ -1,6 +1,7 @@
 import regex as re
 import string
 import os
+import numpy as np
 
 STOPWORDS = ['i', 'me', 'my', 'myself', 'we', 'our', 'ours', 'ourselves', 'you',
 "you're", "you've", "you'll", "you'd", 'your', 'yours', 'yourself', 'yourselves',
@@ -26,15 +27,7 @@ TOPICWORDS = ["no","not","nor",'ain', 'aren', "aren't", 'couldn', "couldn't",
 , "denied", "b.i.d.", "t.i.d.", "p.r.n.", "p.o.", "p.r.", "a.m.", "a.s.", "c.c.",
 "n.p.o.", "o.d.", "o.s.", "o.u.", "q.s.", "q.o.d.","t.i.d.", "t.i.n."]
 
-JOINERWORDS = [
-"of",
-"in",
-"to",
-"on",
-"than",
-"at"
-]
-
+JOINERWORDS = ["of", "in", "to", "on","than","at"]
 
 def appendJoiners(merged, ranges, text, joinerwords):
     if len(joinerwords) == 0:
@@ -50,8 +43,6 @@ def appendJoiners(merged, ranges, text, joinerwords):
             start = merged[i-1][1]
         if i < len(merged)-1:
             end = merged[i+1][0]
-
-        # must have text before and after it...
         if start != merged[i][0] and end != merged[i][1]:
             ranges.append((start,end, 'token'))
     return ranges
@@ -70,11 +61,9 @@ def mergeOverlappingTuples(matches):
         res.append((matches[prev][0], matches[-1][1], 'noise'))
     return res
 
-
 def figureOutRegex(stopwords, topicwords, size=2):
     punc = [x for x in string.punctuation]
     regex = '\n|[\s' + "|\\".join(punc)+ ']{'+ str(size)+',}'
-
     topics = "|".join(["(\s|^)"+stop+"(\s|$)" for stop in topicwords])
     stops = "|".join(["(\s|^)"+stop+"(\s|$)" for stop in stopwords])
 
@@ -82,12 +71,9 @@ def figureOutRegex(stopwords, topicwords, size=2):
         regex = topics +"|" + regex
     if len(stops) != 0:
         regex = stops+'|'+regex
-
-    regex = "("+regex+")"
-    prog = re.compile(regex)
+    prog = re.compile("("+regex+")")
     return prog, topics
 
-
 def appendMatches(matches, ranges, text):
     prev = 0
     for match in matches:
@@ -105,12 +91,11 @@ def getRangesFromIter(matches):
     return ranges
 
 def appendTopics(matches, ranges, text, topics):
-
     if len(topics) == 0:
         return ranges, matches
     matchret = []
     for match in matches:
-        if re.match("("+topics+")", text[match[0]:match[1]]):
+        if re.match("("+topics+")", text[match[0]:match[1]].lower()):
             start = match[0]
             end = match[1]
             if text[match[0]].isspace():
@@ -122,54 +107,132 @@ def appendTopics(matches, ranges, text, topics):
         matchret.append(match)
     return ranges, matchret
 
+def getNoiseProfileLine(tokens, i):
+    if i == -1:
+        return ""
+    line = tokens[i]["token"]
+    if line.find(":") != -1:
+        line = line[line.find(":")+1:]
+    done = False
+    for j in range(i+1, len(tokens)):
+        if not done and tokens[j]["tokenType"] != "noise":
+            line = line + "<TOKEN>"
+        if tokens[j]["tokenType"] == "noise":
+            if done and "\n" not in tokens[j]["token"]:
+                continue
+            done = True
+            line = line + tokens[j]["token"]
+            if "\n" in tokens[j]["token"]:
+                return line
+    return line
 
-def wildgram(text, stopwords=STOPWORDS, topicwords=TOPICWORDS, include1gram=True, joinerwords=JOINERWORDS, returnNoise=True):
+def assignParents(tokens):
+    ranks = {}
+    ranks["root"] = "root"
+    ranks["\n"] = [":\n", "root"]
+    ranks[":\n"] = ["\n", "root"]
+    ranks[": "] = ["\n", ". ", ";"]
+    ranks[". "] = ["\n", ":\n"]
+    ranks[","] = [";", ". ", "\n", ":\n","|", "-", ": "]
+    ranks[";"] = [". ", "\n",":\n"]
+    ranks["-"] = [";", ". ", "\n",":\n", ",", "|"]
+    ranks["|"] = [";", ". ", "\n",":\n", ",", "|"]
+    ranks["and"] = [";", ". ", "\n",":\n", "|", "-"]
+    ranks["or"] = [";", ". ", "\n",":\n", "|", "-"]
+    ranks["by"] = [";", ". ", "\n",":\n", ",", "|", "-"]
+    ranks["in"] = [";", ". ", "\n",":\n", ",", "|", "-"]
+    ranks["on"] = [";", ". ", "\n",":\n", ",", "|", "-"]
+    ranks["was"] = [";", ". ", "\n", ":\n",",", "|", "-"]
+    ranks["is"] = [";", ". ", "\n",":\n", ",", "|", "-"]
+
+    order = [":\n", "\n", ". ", ";", ": ", ",", "-", "and", "or", "|", "by", "in", "on", "was", "is"]
+
+    lastKnown = {}
+    lastKnownKeys = order + ["root", "newLineAny"]
+    for key in lastKnownKeys:
+        lastKnown[key] = -1
+
+    for i in range(len(tokens)):
+        if tokens[i]["tokenType"] != "noise":
+            prev = np.amax(np.array(list(lastKnown.values())))
+            tokens[i]["parent"] = prev
+            continue
+        snip = tokens[i]["token"]
+        done = False
+        for ord in order:
+            if ord not in snip:
+                continue
+
+            if ord in snip:
+                prev = np.amax(np.array(list([lastKnown[r] for r in ranks[ord]])))
+
+                ## if its a plain new line, determining if it should return to root.
+                if "\n" in snip and ":" not in snip:
+                    line = getNoiseProfileLine(tokens, i)
+                    line2 = getNoiseProfileLine(tokens, lastKnown["newLineAny"])
+                    if line.strip() != line2.strip():
+                        prev = -1
+
+                if "\n" in snip:
+                    lastKnown["newLineAny"] = i
+                tokens[i]["parent"] = prev
+                lastKnown[ord] = i
+                done = True
+                break
+        if done:
+            continue
+        prev = np.max(np.array(list(lastKnown.values())))
+        tokens[i]["parent"] = prev
+    return tokens
+
+def createRangesFromProg(prog, topics, text, ranges, noise, joinerwords):
+    matches = getRangesFromIter(prog.finditer(text.lower(),overlapped=True))
+    ranges = appendMatches(matches, ranges, text)
+    if len(topics) != 0:
+        ranges, matches = appendTopics(matches, ranges, text, topics)
+    merged = mergeOverlappingTuples(matches)
+    ranges =appendJoiners(merged, ranges, text, joinerwords)
+    noise = matches + noise
+    return noise, ranges
+
+def wildgram(text, stopwords=STOPWORDS, topicwords=TOPICWORDS, include1gram=True, joinerwords=JOINERWORDS, returnNoise=True, includeParent=True):
     # corner case for inappropriate input
     if not isinstance(text, str):
         raise Exception("What you just gave wildgram isn't a string, mate.")
+    if not returnNoise and includeParent:
+        raise Exception("Parent is based on noise index, you need to set returnNoise to True in order to have includeParent be True. Otherwise set both to False.")
+
     if text.isspace() and not returnNoise:
         return []
-
     if text.isspace() and returnNoise:
-        return [{"startIndex": 0, "endIndex":len(text), "token":text, "tokenType": "noise"}]
+        return [{"startIndex": 0, "endIndex":len(text), "token":text, "tokenType": "noise", "index": 0}]
 
     prog,topics = figureOutRegex(stopwords, topicwords)
     ranges = []
-    matches = getRangesFromIter(prog.finditer(text.lower(),overlapped=True))
-    ranges = appendMatches(matches, ranges, text)
-    ranges, matches = appendTopics(matches, ranges, text, topics)
-    merged = mergeOverlappingTuples(matches)
-    ranges =appendJoiners(merged, ranges, text, joinerwords)
+    noise = []
+    noise, ranges = createRangesFromProg(prog, topics, text, ranges, noise, joinerwords)
 
-    noise = matches
     if include1gram:
         prog1gram,_ = figureOutRegex(stopwords,[], 1)
-        matches = getRangesFromIter(prog1gram.finditer(text.lower(), overlapped=True))
-        noise = noise + matches
-        ranges = appendMatches(matches, ranges, text)
-        merged = mergeOverlappingTuples(matches)
-        ranges = appendJoiners(merged, ranges, text, joinerwords)
+        noise, ranges = createRangesFromProg(prog1gram, [], text, ranges, noise, joinerwords)
 
     if returnNoise:
         ranges = ranges + noise
-    ranges =sorted(list(set(ranges)), key=lambda x: (x[0], x[1]))
 
+    ranges =sorted(list(set(ranges)), key=lambda x: (x[0], x[1]))
     ret = []
     for r in ranges:
-        app = {"startIndex": r[0], "endIndex": r[1], "token": text[r[0]:r[1]], "tokenType": r[2]}
-        if len(ret) == 0:
-            ret.append(app)
-            continue
-        if ret[-1]["tokenType"] != "noise":
-            ret.append(app)
-            continue
-        if r[2] != "noise":
+        app = {"startIndex": r[0], "endIndex": r[1], "token": text[r[0]:r[1]], "tokenType": r[2], "index": len(ret)}
+        if len(ret) == 0 or ret[-1]["tokenType"] != "noise" or r[2] != "noise":
             ret.append(app)
             continue
         if ret[-1]["endIndex"] >= r[0]:
             if ret[-1]["endIndex"] < r[1]:
-                ret[-1] = {"startIndex": ret[-1]["startIndex"], "endIndex": r[1], "token": text[ret[-1]["startIndex"]:r[1]], "tokenType": "noise"}
+                ret[-1] = {"startIndex": ret[-1]["startIndex"], "endIndex": r[1], "token": text[ret[-1]["startIndex"]:r[1]], "tokenType": "noise", "index": len(ret)-1}
             continue
         ret.append(app)
 
+    if includeParent:
+        ret = assignParents(ret)
+
     return ret
